import numpy as np
import pandas as pd
import LRFtools as LRF
import gc

import scipy.stats as st
from scipy.optimize import curve_fit

import matplotlib.pyplot as plt
import matplotlib as mpl
import seaborn as sns
mpl.rcParams['figure.figsize'] = [7,8]
mpl.rcParams['figure.dpi'] = 80
mpl.rcParams['savefig.dpi'] = 200

mpl.rcParams['font.size'] = 17
mpl.rcParams['legend.fontsize'] = 'large'
mpl.rcParams['figure.titlesize'] = 'medium'
mpl.rcParams['lines.linewidth'] = 2.5
mpl.rcParams['lines.markersize'] = 10

import seaborn as sns
sns.set_context("poster", rc={"font.size":15,"axes.titlesize":15,"axes.labelsize":15})

p1s, p2s = [], []

# Load the models
def load_all():

	# Load demographic data for all wards
	df_population = pd.read_csv('data/pop_estimates_2016_oxfordshire.tsv', sep='\t')

	df_population['population'] = df_population['All Ages']
	new_cols = list(df_population.columns)
	new_cols[0] = 'CODE'
	new_cols[1] = 'NAME'
	df_population.columns = new_cols

	# Load traffic data
	#df_all_features = pd.read_csv('OpenStreetMap/all_features_per_ward_oxfordshire.tsv', sep='\t')

	# Load all wards
	df_wards = LRF.load_wards()
	wardcodes = df_wards.CODE.values

	# Load traffic data
	df_Tij_data = pd.read_csv('data/df_total_with_triptime.csv', index_col=0)
	df_Tij_data['sij']		= df_Tij_data['new_sij']
	df_Tij_data['meanvolume'] = df_Tij_data['Tij_data']

	df_Tij_data = df_Tij_data.filter(
			['origin', 'destination', 'distance', 'sij', 'Ti', 'meanvolume',
			 'amenity', 'building', 'landuse', 'leisure',
			 'office', 'shop', 'sport', 'avg_scenicness',
		   'sij_amenity', 'sij_building', 'sij_landuse', 'sij_leisure',
		   'sij_office', 'sij_shop', 'sij_sport', 'sij_avg_scenicness',
			 'pop_orig', 'pop_dest', 'old_distance', 'old_sij', 'new_sij',
		   'amenity_origin', 'building_origin', 'landuse_origin', 'leisure_origin',
		   'office_origin', 'shop_origin', 'sport_origin', 'avg_scenicness_origin'
		   ])

	# Reduce models

	thresh1 = 0 # bottom cutoff
	cutoff1 = thresh1 #* pop_traffic/pop_oxfordshire
	#cutoff2 = sorted(df_Tij_data.meanvolume)[ int(0.99*len(df_Tij_data)) ]

 	# no top cutoff
	thresh2 = np.infty
	cutoff2 = np.infty

	mask = (df_Tij_data.meanvolume > cutoff1) & (df_Tij_data.meanvolume < cutoff2)
	df_Tij_data_filtered = df_Tij_data[mask]
	df_Tij_data_filtered_2 = df_Tij_data.copy()
	df_Tij_data_filtered_2['meanvolume'] = [ i if ( (i > cutoff1) & (i < cutoff2) ) else 0 \
											for i in df_Tij_data_filtered_2['meanvolume'] ]

	print('Reduced to %d trips' % len(df_Tij_data_filtered))

	print("Loaded all data")

	loaded_data = { 'df_Tij_data_filtered':df_Tij_data_filtered,
					'df_Tij_data_filtered_2':df_Tij_data_filtered_2,
					'thresh1':thresh1, 'thresh2':thresh2,
					'df_wards':df_wards, 'df_population':df_population,
					'df_Tij_data':df_Tij_data }

	return loaded_data


def fit_all(loaded_data, variation=-1):
	df_wards = loaded_data['df_wards']
	df_population = loaded_data['df_population']
	thresh1 = loaded_data['thresh1']
	thresh2 = loaded_data['thresh2']
	df_Tij_data = loaded_data['df_Tij_data'].copy()
	df = loaded_data['df_Tij_data_filtered'].copy()
	df_Tij_data_filtered_2 = loaded_data['df_Tij_data_filtered_2'].copy()
	global p1s, p2s

	# I want to iterate over:
	# (2) with or without the sij modification
	# (2) with travel time instead of distance or not
	# (4) possible mobility models

	from itertools import product
	l = range(2)
	triples = list(product(l,l,l))

	if variation is not -1:
		triples = [ triples[variation] ]

	for duple in product(l,l):
		triple = (duple[0],0,duple[1])

		with_sij_mod  = triple[0]
		with_OSM_mod  = triple[1]
		with_duration = triple[2]

		print('with_sij_mod:', with_sij_mod>0)
		print('with_OSM_mod:', with_OSM_mod>0)
		print('with_duration:',with_duration>0)

		if with_OSM_mod:
			OSM_keys = [ 'amenity', 'building', 'landuse', 'leisure', 'office', 'shop', 'sport', 'avg_scenicness' ]
			#OSM_keys = [ 'building', 'landuse', 'avg_scenicness' ]
			sij_OSM_keys = [ 'sij_'+tag for tag in OSM_keys ]
			OSM_values	 = [ df[tag].values  for tag in OSM_keys ]
			sij_OSM_values = [ df[tag].values  for tag in sij_OSM_keys ]
		else:
			list_tags = []
			OSM_values	 = []
			sij_OSM_values = []


		if not with_duration:
			df['distance'] = df['old_distance']

		# Define fitting variables
		pop_oxfordshire = df_population.population.sum()
		pop_traffic	 = df.meanvolume.sum()
		pop_orig  = df['pop_orig'].values
		pop_dest  = df['pop_dest'].values
		distance  = df['distance'].values
		Tij	   = df['meanvolume'].values
		sij	   = df['sij'].values
		Ti		= df['Ti'].values

		N_FIT_ITERATIONS = 200
		#"""

		print("\nFitting Model 1: gravity, power function on distance")
		# Model 1: gravity, power function on distance

		min_NRMSE_final = np.infty
		y_final		 = np.ones_like(Tij)

		gc.collect()
		for iterations in range(N_FIT_ITERATIONS):
			try:
				p1s, p2s = [], []

				def grav_pow(inputs, *params):
					pop_orig, pop_dest, distance, sij, Ti, pops = inputs[:6]
					pop_oxfordshire, pop_traffic = pops
					OSM_values = inputs[6:]

					grav_power = params[0]
					c = params[1:]
					attractiveness = c[0] * pop_dest

					if np.any(attractiveness<0):
						return -1e10
					else:
						# Gravity + power function
						log_Tij = np.log10(pop_orig) + np.log10(attractiveness)
						log_Tij -= grav_power*np.log10(distance)

						# Normalisation
						log_Tij += (np.log10(pop_oxfordshire) - np.log10(pop_traffic))

						# Linear OSM modification
						if len(OSM_values) != 0:
							# Linear modification
							#log_Tij += np.log10(sum( ci*vi/pop_dest for ci,vi in zip(c[1:],OSM_values) ))
							# Exponential OSM modification
							#log_Tij += np.log10(np.e)*sum( ci*vi/pop_dest for ci,vi in zip(c[1:],OSM_values) )
							# Syrian modification
							log_Tij += ( distance / sum( ci*vi/pop_dest for ci,vi in zip(c[1:],OSM_values) ))


						# Apply cutoff
						m = min(log_y_data)
						#log_Tij = np.array([ i if i>m else m for i in log_Tij ])

						global p1s, p2s
						p1s += [ list(params) + [0]*4 ]
						p2s += [ (log_Tij-log_y_data)**2 ]		

						return log_Tij


				p0 = [0.0]*2 + [ 0.0 for _ in OSM_values ]
				p0 = np.random.rand(len(p0))
				func = grav_pow

				pops   = [ pop_oxfordshire, pop_traffic ]
				x_data = [ pop_orig, pop_dest, distance, sij, Ti, pops ] + OSM_values
				x_data = np.array([ np.array(x, dtype='float64') for x in x_data ])
				y_data = np.array(Tij.copy(), dtype='float64')
				log_y_data = np.log10(y_data)

				fitParams, fitCovariances = curve_fit(func, x_data, log_y_data, p0=p0, maxfev=30000)
				y_grav_pow1 = func(x_data, *fitParams)

				#print('%d iterations for part 1' % len(p1s), end=', ')
				
				if with_sij_mod:
					def grav_pow_mod(inputs, *params):
						pop_orig, pop_dest, distance, sij, Ti, pops, y_grav_pow = inputs[:7]
						pop_oxfordshire, pop_traffic = pops
						OSM_values = inputs[7:]

						A, B = params

						if (A<0) or (B<0):
							return -1e20

						else:
							# Gravity + power function
							Tij = 10**y_grav_pow

							# sij modification
							Tij *= ( A * np.exp(-sij/(B*100000)) + 1 )
							#Tij *= ( pop_orig ** C ) * ( pop_dest ** D )

							log_Tij = np.log10(Tij)
							global p1s, p2s
							p1s += [ p1s[-1][:-4] + list(params) ]
							p2s += [ ((log_Tij-log_y_data)**2)/sum(log_y_data) ]

							return Tij

					p0 = [0.0]*2
					p0 = np.random.rand(len(p0))
					func = grav_pow_mod

					pops   = [ pop_oxfordshire, pop_traffic ]
					x_data = [ pop_orig, pop_dest, distance, sij, Ti, pops, y_grav_pow1 ] + OSM_values
					x_data = np.array([ np.array(x, dtype='float64') for x in x_data ])
					y_data = np.array(Tij.copy(), dtype='float64')
					log_y_data = np.log10(y_data)

					fitParams, fitCovariances = curve_fit(func, x_data, y_data, p0=p0, maxfev=30000)
					y_grav_pow = func(x_data, *fitParams)

					y_grav_pow = np.log10(y_grav_pow)

				else:
					y_grav_pow = y_grav_pow1

				#print('%d iterations for part 2' % len(p1s), end=', ')

				NRMSE_final = sum(p2s[-1])
				#print('NRMSE = %g' % NRMSE_final)

				if NRMSE_final < min_NRMSE_final:
					y_final = y_grav_pow.copy()
					min_NRMSE_final = NRMSE_final
					print('iteration %d: NRMSE = %g' % (iterations, min_NRMSE_final) )

			except RuntimeError:
				print('Iteration %d, RuntimeError' % iterations)
				continue
				
		y_grav_pow = y_final.copy()
		print('Final NRMSE = %g' % (min_NRMSE_final) )		
		

		# Model 2: gravity, exponential function on distance
		print("\nFitting Model 2: gravity, exponential function on distance")

		min_NRMSE_final = np.infty
		y_final		 = np.ones_like(Tij)

		gc.collect()
		for iterations in range(N_FIT_ITERATIONS):
			try:
				p1s, p2s = [], []

				def grav_exp(inputs, *params):
					pop_orig, pop_dest, distance, sij, Ti, pops = inputs[:6]
					pop_oxfordshire, pop_traffic = pops
					OSM_values = inputs[6:]

					grav_exponent = params[0]
					c = params[1:]
					attractiveness = c[0] * pop_dest

					if np.any(attractiveness<0):
						return -1e10
					else:
						# Gravity + power function
						log_Tij = np.log10(pop_orig) + np.log10(attractiveness)
						log_Tij -= grav_exponent*distance

						# Normalisation
						log_Tij += (np.log10(pop_oxfordshire) - np.log10(pop_traffic))

						# Linear OSM modification
						if len(OSM_values) != 0:
							# Linear modification
							#log_Tij += np.log10(sum( ci*vi/pop_dest for ci,vi in zip(c[1:],OSM_values) ))
							# Exponential OSM modification
							#log_Tij += np.log10(np.e)*sum( ci*vi/pop_dest for ci,vi in zip(c[1:],OSM_values) )
							# Syrian modification
							log_Tij += ( distance / sum( ci*vi/pop_dest for ci,vi in zip(c[1:],OSM_values) ))

						# Exponential OSM modification
						#log_Tij += np.log10(np.e)*sum( ci*vi/pop_dest for ci,vi in zip(c[1:],OSM_values) )

						# Apply cutoff
						m = min(log_y_data)
						#log_Tij = np.array([ i if i>m else m for i in log_Tij ])

						global p1s, p2s
						p1s += [ list(params) + [0]*4 ]
						p2s += [ (log_Tij-log_y_data)**2 ]

						return log_Tij

				p0 = [0.0]*2 + [ 0.0 for _ in OSM_values ]
				p0 = np.random.rand(len(p0))
				func = grav_exp

				pops   = [ pop_oxfordshire, pop_traffic ]
				x_data = [ pop_orig, pop_dest, distance, sij, Ti, pops ] + OSM_values
				x_data = np.array([ np.array(x, dtype='float64') for x in x_data ])
				y_data = np.array(Tij.copy(), dtype='float64')
				log_y_data = np.log10(y_data)

				fitParams, fitCovariances = curve_fit(func, x_data, log_y_data, p0=p0, maxfev=30000)
				y_grav_exp1 = func(x_data, *fitParams)

				#print('%d iterations for part 1' % len(p1s), end=', ')

				if with_sij_mod:
					def grav_exp_mod(inputs, *params):
						pop_orig, pop_dest, distance, sij, Ti, pops, y_grav_exp = inputs[:7]
						pop_oxfordshire, pop_traffic = pops
						OSM_values = inputs[7:]

						A, B = params

						if (A<0) or (B<0):
							return -1e10

						else:
							# Gravity + power function
							Tij = 10**y_grav_exp

							# sij modification
							Tij *= ( A * np.exp(-sij/(B*100000)) + 1 )
							#Tij *= ( pop_orig ** C ) * ( pop_dest ** D )

							log_Tij = np.log10(Tij)
							global p1s, p2s
							p1s += [ p1s[-1][:-4] + list(params) ]
							p2s += [ ((log_Tij-log_y_data)**2)/sum(log_y_data) ]

							return Tij


					p0 = [0.0]*2
					p0 = np.random.rand(len(p0))
					func = grav_exp_mod

					pops   = [ pop_oxfordshire, pop_traffic ]
					x_data = [ pop_orig, pop_dest, distance, sij, Ti, pops, y_grav_exp1 ] + OSM_values
					x_data = np.array([ np.array(x, dtype='float64') for x in x_data ])
					y_data = np.array(Tij.copy(), dtype='float64')
					log_y_data = np.log10(y_data)

					fitParams, fitCovariances = curve_fit(func, x_data, y_data, p0=p0, maxfev=30000)
					y_grav_exp = func(x_data, *fitParams)

					y_grav_exp = np.log10(y_grav_exp)

				else:
					y_grav_exp = y_grav_exp1

				#print('%d iterations for part 2' % len(p1s), end=', ')

				NRMSE_final = sum(p2s[-1])
				#print('NRMSE = %g' % NRMSE_final)

				if NRMSE_final < min_NRMSE_final:
					y_final = y_grav_exp.copy()
					min_NRMSE_final = NRMSE_final
					print('iteration %d: NRMSE = %g' % (iterations, min_NRMSE_final) )
					
			except RuntimeError:
				print('Iteration %d, RuntimeError' % iterations)
				continue

		y_grav_exp = y_final.copy()
		print('Final NRMSE = %g' % (min_NRMSE_final) )

		print("\nFitting Model 3: gravity, power function on pops and distance")
		# Model 3: gravity, power function on pops and distance

		min_NRMSE_final = np.infty
		y_final		 = np.ones_like(Tij)

		gc.collect()
		print('len(OSM_values) =',len(OSM_values))
		for iterations in range(N_FIT_ITERATIONS):
			try:
				p1s, p2s = [], []

				def grav_pow2(inputs, *params):
					pop_orig, pop_dest, distance, sij, Ti, pops = inputs[:6]
					pop_oxfordshire, pop_traffic = pops
					OSM_values = inputs[6:]

					grav_power = params[0]
					G = params[1]
					c = params[2:]
					attractiveness = c[0]*np.log10(pop_dest) + G
					repulsiveness  = c[1]*np.log10(pop_orig)

					if np.any(attractiveness<0):
						return -1e10
					else:
						# Gravity + power function
						log_Tij =  attractiveness + repulsiveness
						log_Tij -= grav_power*np.log10(distance)

						# Normalisation
						log_Tij += (np.log10(pop_oxfordshire) - np.log10(pop_traffic))

						# Linear OSM modification
						if len(OSM_values) != 0:
							# Linear modification
							#log_Tij += np.log10(sum( ci*vi/pop_dest for ci,vi in zip(c[1:],OSM_values) ))
							# Exponential OSM modification
							#log_Tij += np.log10(np.e)*sum( ci*vi/pop_dest for ci,vi in zip(c[1:],OSM_values) )
							# Syrian modification
							log_Tij += ( distance / sum( ci*vi/pop_dest for ci,vi in zip(c[1:],OSM_values) ))

						# Exponential OSM modification
						#log_Tij += np.log10(np.e)*sum( ci*vi/pop_dest for ci,vi in zip(c[1:],OSM_values) )

						# Apply cutoff
						m = min(log_y_data)
						#log_Tij = np.array([ i if i>m else m for i in log_Tij ])	

						global p1s, p2s
						p1s += [ list(params) + [0]*4 ]
						p2s += [ (log_Tij-log_y_data)**2 ]

						return log_Tij

				p0 = [0.0]*4 + [ 0.0 for _ in OSM_values ]
				p0 = np.random.rand(len(p0))
				func = grav_pow2

				pops   = [ pop_oxfordshire, pop_traffic ]
				x_data = [ pop_orig, pop_dest, distance, sij, Ti, pops ] + OSM_values
				x_data = np.array([ np.array(x, dtype='float64') for x in x_data ])
				y_data = np.array(Tij.copy(), dtype='float64')
				log_y_data = np.log10(y_data)

				fitParams, fitCovariances = curve_fit(func, x_data, log_y_data, p0=p0, maxfev=30000)
				y_grav_pow1 = func(x_data, *fitParams)

				print('Part 1, iteration %d: %d iterations' % (iterations,len(p1s)))

				if with_sij_mod:
					def grav_pow_mod2(inputs, *params):
						pop_orig, pop_dest, distance, sij, Ti, pops, y_grav_pow = inputs[:7]
						pop_oxfordshire, pop_traffic = pops
						OSM_values = inputs[7:]

						A, B = params

						if (A<0) or (B<0):
							return -1e10

						else:
							# Gravity + power function
							Tij = 10**y_grav_pow

							# sij modification
							Tij *= ( A * np.exp(-sij/(B*100000)) + 1 )
							#Tij *= ( pop_orig ** C ) * ( pop_dest ** D )

							log_Tij = np.log10(Tij)
							global p1s, p2s
							p1s += [ p1s[-1][:-4] + list(params) ]
							p2s += [ ((log_Tij-log_y_data)**2)/sum(log_y_data) ]

							return Tij


					p0 = [0.0]*2
					p0 = np.random.rand(len(p0))
					func = grav_pow_mod2

					pops   = [ pop_oxfordshire, pop_traffic ]
					x_data = [ pop_orig, pop_dest, distance, sij, Ti, pops, y_grav_pow1 ] + OSM_values
					x_data = np.array([ np.array(x, dtype='float64') for x in x_data ])
					y_data = np.array(Tij.copy(), dtype='float64')
					log_y_data = np.log10(y_data)

					fitParams, fitCovariances = curve_fit(func, x_data, y_data, p0=p0, maxfev=30000)
					y_grav_pow2 = func(x_data, *fitParams)

					y_grav_pow2 = np.log10(y_grav_pow2)

				else:
					y_grav_pow2 = y_grav_pow1

				#print('%d iterations for part 2' % len(p1s), end=', ')

				NRMSE_final = sum(p2s[-1])
				#print('NRMSE = %g' % NRMSE_final)

				if NRMSE_final < min_NRMSE_final:
					y_final = y_grav_pow2.copy()
					min_NRMSE_final = NRMSE_final
					print('iteration %d: NRMSE = %g' % (iterations, min_NRMSE_final) )

			except RuntimeError:
				print('Part 2, iteration %d, RuntimeError' % iterations)
				continue
					
		y_grav_pow2 = y_final.copy()
		print('Final NRMSE = %g' % ( min_NRMSE_final) )

		print("\nFitting Model 4: gravity, power function on pops, exp on distance")
		# Model 4: gravity, power function on pops, exp on distance

		min_NRMSE_final = np.infty
		y_final		 = np.ones_like(Tij)

		gc.collect()
		for iterations in range(N_FIT_ITERATIONS):
			try:
				p1s, p2s = [], []

				def grav_exp2(inputs, *params):
					pop_orig, pop_dest, distance, sij, Ti, pops = inputs[:6]
					pop_oxfordshire, pop_traffic = pops
					OSM_values = inputs[6:]

					grav_exponent = params[0]
					G = params[1]
					c = params[2:]
					attractiveness = c[0]*np.log10(pop_dest) + G
					repulsiveness  = c[1]*np.log10(pop_orig)

					if np.any(attractiveness<0):
						return -1e10
					else:
						# Gravity + power function
						log_Tij =  attractiveness + repulsiveness
						log_Tij -= grav_exponent*distance

						# Normalisation
						log_Tij += (np.log10(pop_oxfordshire) - np.log10(pop_traffic))

						# Linear OSM modification
						if len(OSM_values) != 0:
							# Linear modification
							#log_Tij += np.log10(sum( ci*vi/pop_dest for ci,vi in zip(c[1:],OSM_values) ))
							# Exponential OSM modification
							#log_Tij += np.log10(np.e)*sum( ci*vi/pop_dest for ci,vi in zip(c[1:],OSM_values) )
							# Syrian modification
							log_Tij += ( distance / sum( ci*vi/pop_dest for ci,vi in zip(c[1:],OSM_values) ))

						# Exponential OSM modification
						#log_Tij += np.log10(np.e)*sum( ci*vi/pop_dest for ci,vi in zip(c[1:],OSM_values) )

						# Apply cutoff
						m = min(log_y_data)
						#log_Tij = np.array([ i if i>m else m for i in log_Tij ])		

						global p1s, p2s
						p1s += [ list(params) + [0]*4 ]
						p2s += [ (log_Tij-log_y_data)**2 ]

						return log_Tij

				p0 = [0.0]*4 + [ 0.0 for _ in OSM_values ]
				p0 = np.random.rand(len(p0))
				func = grav_exp2

				pops   = [ pop_oxfordshire, pop_traffic ]
				x_data = [ pop_orig, pop_dest, distance, sij, Ti, pops ] + OSM_values
				x_data = np.array([ np.array(x, dtype='float64') for x in x_data ])
				y_data = np.array(Tij.copy(), dtype='float64')
				log_y_data = np.log10(y_data)

				fitParams, fitCovariances = curve_fit(func, x_data, log_y_data, p0=p0, maxfev=30000)
				y_grav_exp1 = func(x_data, *fitParams)

				#print('%d iterations for part 1' % len(p1s), end=', ')

				if with_sij_mod:
					def grav_exp_mod2(inputs, *params):
						pop_orig, pop_dest, distance, sij, Ti, pops, y_grav_exp = inputs[:7]
						pop_oxfordshire, pop_traffic = pops
						OSM_values = inputs[7:]

						A, B = params

						if (A<0) or (B<0):
							return -1e10

						else:
							# Gravity + power function
							Tij = 10**y_grav_exp

							# sij modification
							Tij *= ( A * np.exp(-sij/(B*100000)) + 1 )
							#Tij *= ( pop_orig ** C ) * ( pop_dest ** D )

							log_Tij = np.log10(Tij)
							global p1s, p2s
							p1s += [ p1s[-1][:-4] + list(params) ]
							p2s += [ ((log_Tij-log_y_data)**2)/sum(log_y_data) ]

							return Tij


					p0 = [0.0]*2
					p0 = np.random.rand(len(p0))
					func = grav_exp_mod2

					pops   = [ pop_oxfordshire, pop_traffic ]
					x_data = [ pop_orig, pop_dest, distance, sij, Ti, pops, y_grav_exp1 ] + OSM_values
					x_data = np.array([ np.array(x, dtype='float64') for x in x_data ])
					y_data = np.array(Tij.copy(), dtype='float64')
					log_y_data = np.log10(y_data)

					fitParams, fitCovariances = curve_fit(func, x_data, y_data, p0=p0, maxfev=30000)
					y_grav_exp2 = func(x_data, *fitParams)

					y_grav_exp2 = np.log10(y_grav_exp2)

				else:
					y_grav_exp2 = y_grav_exp1

				#print('%d iterations for part 2' % len(p1s), end=', ')

				NRMSE_final = sum(p2s[-1])
				#print('NRMSE = %g' % NRMSE_final)

				if NRMSE_final < min_NRMSE_final:
					y_final = y_grav_exp2.copy()
					min_NRMSE_final = NRMSE_final
					print('iteration %d: NRMSE = %g' % (iterations, min_NRMSE_final) )

			except RuntimeError:
				print('Iteration %d, RuntimeError' % iterations)
				continue
				
		y_grav_exp2 = y_final.copy()
		print('Final NRMSE = %g' % ( min_NRMSE_final) )

		print("\nFitting Model 5: original radiation model")
		# Model 5: original radiation model

		min_NRMSE_final = np.infty
		y_final		 = np.ones_like(Tij)

		gc.collect()
		for iterations in range(N_FIT_ITERATIONS):
			try:
				p1s, p2s = [], []

				def rad_ori(inputs,  *params):
					idx_OSM_values = 6
					idx_sij_OSM_values = idx_OSM_values + len(params[1:])

					pop_orig, pop_dest, distance, sij, Ti, pops = inputs[:idx_OSM_values]
					pop_oxfordshire, pop_traffic = pops   
					OSM_values	 = inputs[idx_OSM_values:idx_sij_OSM_values]
					sij_OSM_values = inputs[idx_sij_OSM_values:]

					c = params
					attractiveness = c[0]*pop_dest 
					sij_all		= c[0]*sij

					if np.any(attractiveness<0) or np.any(sij_all<0):
						return -1e10
					else:
						# numerator
						log_Tij = np.log10(Ti) + np.log10(pop_orig) + np.log10(attractiveness)

						# denominator
						log_Tij -= np.log10(pop_orig + sij_all)
						log_Tij -= np.log10(pop_orig + attractiveness + sij_all)

						# scaling
						log_Tij += (np.log10(pop_oxfordshire) - np.log10(pop_traffic))

						# Linear OSM modification
						if len(OSM_values) != 0:
							# Linear modification
							#log_Tij += np.log10(sum( ci*vi/pop_dest for ci,vi in zip(c[1:],OSM_values) ))
							# Exponential OSM modification
							#log_Tij += np.log10(np.e)*sum( ci*vi/pop_dest for ci,vi in zip(c[1:],OSM_values) )
							# Syrian modification
							log_Tij += ( distance / sum( ci*vi/pop_dest for ci,vi in zip(c[1:],OSM_values) ))

						# Exponential OSM modification
						#log_Tij += np.log10(np.e)*sum( ci*vi/pop_dest for ci,vi in zip(c[1:],OSM_values) )

						# Apply cutoff
						m = min(log_y_data)
						#log_Tij = np.array([ i if i>m else m for i in log_Tij ])		

						global p1s, p2s
						p1s += [ list(params) + [0]*4 ]
						p2s += [ (log_Tij-log_y_data)**2 ]

						return log_Tij

				p0 = [0.0] + [ 0.0 for _ in OSM_values ]
				p0 = np.random.rand(len(p0))
				func = rad_ori

				pops   = [ pop_oxfordshire, pop_traffic ]
				x_data = [ pop_orig, pop_dest, distance, sij, Ti, pops ] + OSM_values + sij_OSM_values
				x_data = np.array([ np.array(x, dtype='float64') for x in x_data ])
				y_data = np.array(Tij.copy(), dtype='float64')
				log_y_data = np.log10(y_data)

				fitParams, fitCovariances = curve_fit(func, x_data, log_y_data, p0=p0, maxfev=50000)
				y_rad_ori1 = func(x_data, *fitParams)

				#print('%d iterations for part 1' % len(p1s), end=', ')

				if with_sij_mod:
					def rad_ori_mod(inputs, *params):
						pop_orig, pop_dest, distance, sij, Ti, pops, y_rad_ori = inputs[:7]
						pop_oxfordshire, pop_traffic = pops
						OSM_values = inputs[7:]

						A, B = params

						if (A<0) or (B<0):
							return -1e10

						else:
							# Non-modified model
							Tij = 10**y_rad_ori

							# sij modification
							Tij *= ( A * np.exp(-sij/(B*100000)) + 1 )
							#Tij *= ( pop_orig ** C ) * ( pop_dest ** D )

							# Apply cutoff
							m = min(y_data)
							#Tij = np.array([ i if i>m else m for i in Tij ])			  

							log_Tij = np.log10(Tij)
							global p1s, p2s
							p1s += [ p1s[-1][:-4] + list(params) ]
							p2s += [ ((log_Tij-log_y_data)**2)/sum(log_y_data) ]

							return Tij


					p0 = [0.0]*2
					p0 = np.random.rand(len(p0))
					func = rad_ori_mod

					pops   = [ pop_oxfordshire, pop_traffic ]
					x_data = [ pop_orig, pop_dest, distance, sij, Ti, pops, y_rad_ori1 ] + OSM_values
					x_data = np.array([ np.array(x, dtype='float64') for x in x_data ])
					y_data = np.array(Tij.copy(), dtype='float64')
					log_y_data = np.log10(y_data)

					fitParams, fitCovariances = curve_fit(func, x_data, y_data, p0=p0, maxfev=30000)
					y_rad_ori = func(x_data, *fitParams)

					y_rad_ori = np.log10(y_rad_ori)

				else:
					y_rad_ori = y_rad_ori1

				#print('%d iterations for part 2' % len(p1s), end=', ')

				NRMSE_final = sum(p2s[-1])
				#print('NRMSE = %g' % NRMSE_final)

				if NRMSE_final < min_NRMSE_final:
					y_final = y_rad_ori.copy()
					min_NRMSE_final = NRMSE_final
					print('iteration %d: NRMSE = %g' % (iterations, min_NRMSE_final) )

			except RuntimeError:
				print('Iteration %d, RuntimeError' % iterations)
				continue

		y_rad_ori = y_final.copy()
		print('Final NRMSE = %g' % ( min_NRMSE_final) )

		print("\nFitting Model 6: radiation model, with one parameter (alpha)")
		# Model 6: radiation model, with one parameter (alpha)

		min_NRMSE_final = np.infty
		y_final		 = np.ones_like(Tij)

		gc.collect()
		for iterations in range(N_FIT_ITERATIONS):
			try:
				p1s, p2s = [], []

				def rad_alpha(inputs,  *params):
					idx_OSM_values = 6
					idx_sij_OSM_values = idx_OSM_values + len(params[1:])

					pop_orig, pop_dest, distance, sij, Ti, pops = inputs[:idx_OSM_values]
					pop_oxfordshire, pop_traffic = pops   
					OSM_values	 = inputs[idx_OSM_values:idx_sij_OSM_values]
					sij_OSM_values = inputs[idx_sij_OSM_values:]

					alpha = params[0]
					c = params[1:]
					attractiveness = c[0]*pop_dest 
					sij_all		= c[0]*sij

					if np.any(attractiveness<0) or np.any(sij_all<0):
						return -1e10
					else:
						# numerator
						log_Tij = np.log10(Ti)
						log_Tij += np.log10((pop_orig + attractiveness + sij_all)**alpha-(pop_orig + sij_all)**alpha )
						log_Tij += alpha*np.log10(pop_orig**alpha + 1)

						# denominator
						log_Tij -= np.log10((pop_orig + sij_all)**alpha + 1)
						log_Tij -= np.log10((pop_orig + attractiveness + sij_all)**alpha +1)

						# scaling
						log_Tij += (np.log10(pop_oxfordshire) - np.log10(pop_traffic))

						# Linear OSM modification
						if len(OSM_values) != 0:
							# Linear modification
							#log_Tij += np.log10(sum( ci*vi/pop_dest for ci,vi in zip(c[1:],OSM_values) ))
							# Exponential OSM modification
							#log_Tij += np.log10(np.e)*sum( ci*vi/pop_dest for ci,vi in zip(c[1:],OSM_values) )
							# Syrian modification
							log_Tij += ( distance / sum( ci*vi/pop_dest for ci,vi in zip(c[1:],OSM_values) ))

						# Exponential OSM modification
						#log_Tij += np.log10(np.e)*sum( ci*vi/pop_dest for ci,vi in zip(c[1:],OSM_values) )

						# Apply cutoff
						m = min(log_y_data)
						#log_Tij = np.array([ i if i>m else m for i in log_Tij ])		

						global p1s, p2s
						p1s += [ list(params) + [0]*4 ]
						p2s += [ (log_Tij-log_y_data)**2 ]

						return log_Tij

				p0 = [0.0]*2 + [ 0.0 for _ in OSM_values ]
				p0 = np.random.rand(len(p0))
				func = rad_alpha

				pops   = [ pop_oxfordshire, pop_traffic ]
				x_data = [ pop_orig, pop_dest, distance, sij, Ti, pops ] + OSM_values + sij_OSM_values
				x_data = np.array([ np.array(x, dtype='float64') for x in x_data ])
				y_data = np.array(Tij.copy(), dtype='float64')
				log_y_data = np.log10(y_data)

				fitParams, fitCovariances = curve_fit(func, x_data, log_y_data, p0=p0, maxfev=50000)
				y_rad_alpha1 = func(x_data, *fitParams)

				#print('%d iterations for part 1' % len(p1s), end=', ')

				if with_sij_mod:
					def rad_alpha_mod(inputs, *params):
						pop_orig, pop_dest, distance, sij, Ti, pops, y_rad_alpha = inputs[:7]
						pop_oxfordshire, pop_traffic = pops
						OSM_values = inputs[7:]

						A, B = params

						if (A<0) or (B<0):
							return -1e10

						else:
							# Non-modified model
							Tij = 10**y_rad_alpha

							# sij modification
							Tij *= ( A * np.exp(-sij/(B*100000)) + 1 )
							#Tij *= ( pop_orig ** C ) * ( pop_dest ** D )

							log_Tij = np.log10(Tij)
							global p1s, p2s
							p1s += [ p1s[-1][:-4] + list(params) ]
							p2s += [ ((log_Tij-log_y_data)**2)/sum(log_y_data) ]

							return Tij


					p0 = [0.0]*2
					p0 = np.random.rand(len(p0))
					func = rad_alpha_mod

					pops   = [ pop_oxfordshire, pop_traffic ]
					x_data = [ pop_orig, pop_dest, distance, sij, Ti, pops, y_rad_alpha1 ] + OSM_values
					x_data = np.array([ np.array(x, dtype='float64') for x in x_data ])
					y_data = np.array(Tij.copy(), dtype='float64')
					log_y_data = np.log10(y_data)

					fitParams, fitCovariances = curve_fit(func, x_data, y_data, p0=p0, maxfev=30000)
					y_rad_alpha = func(x_data, *fitParams)

					y_rad_alpha = np.log10(y_rad_alpha)

				else:
					y_rad_alpha = y_rad_alpha1

				#print('%d iterations for part 2' % len(p1s), end=', ')

				NRMSE_final = sum(p2s[-1])
				#print('NRMSE = %g' % NRMSE_final)

				if NRMSE_final < min_NRMSE_final:
					y_final = y_rad_alpha.copy()
					min_NRMSE_final = NRMSE_final
					print('iteration %d: NRMSE = %g' % (iterations, min_NRMSE_final) )

			except RuntimeError:
				print('Iteration %d, RuntimeError' % iterations)
				continue
					
		y_rad_alpha = y_final.copy()
		print('Final NRMSE = %g' % ( min_NRMSE_final) )		
		#"""

		def reproject(latlonlist):
			latitude  = latlonlist[:,0]
			longitude = latlonlist[:,1]
			# Returns the x & y coordinates in meters using a sinusoidal projection
			from math import pi, cos, radians
			earth_radius = 6371009 # in meters
			lat_dist = pi * earth_radius / 180.0

			y = [lat * lat_dist for lat in latitude]
			x = [long * lat_dist * cos(radians(lat)) 
						for lat, long in zip(latitude, longitude)]
			#return x, y
			return zip(x,y)


		# determine alpha using method in EPJ paper
		# L = mean square root of ward area
		from shapely.geometry import Point, Polygon, shape

		latlonlists = [ np.array(i[0]) for i in df_wards['coordinates'] ]
		ward_polys = [ Polygon(reproject(i)) for i in latlonlists]
		ward_areas = [ shape(poly).area for poly in ward_polys ]
		mean_area_km2 = np.mean(ward_areas)/1000000
		L = np.sqrt(mean_area_km2)

		alpha = np.array([( L/36 )**1.33])			

		print("\nFitting Model 7: radiation model, with alpha already set")
		# Model 7: radiation model, with alpha already set

		min_NRMSE_final = np.infty
		y_final		 = np.ones_like(Tij)

		gc.collect()
		for iterations in range(N_FIT_ITERATIONS):
			try:
				p1s, p2s = [], []

				def rad_alpha2(inputs,  *params):
					idx_OSM_values = 7
					idx_sij_OSM_values = idx_OSM_values + len(params)

					alpha, pop_orig, pop_dest, distance, sij, Ti, pops = inputs[:idx_OSM_values]
					pop_oxfordshire, pop_traffic = pops

					OSM_values	 = inputs[idx_OSM_values:idx_sij_OSM_values]
					sij_OSM_values = inputs[idx_sij_OSM_values:]

					c = params	  
					attractiveness = c[0]*pop_dest 
					sij_all		= c[0]*sij

					if np.any(attractiveness<0) or np.any(sij_all<0):
						return -1e10
					else:
						# numerator
						log_Tij = np.log10(Ti)
						log_Tij += np.log10((pop_orig + attractiveness + sij_all)**alpha-(pop_orig + sij_all)**alpha )
						log_Tij += alpha*np.log10(pop_orig**alpha + 1)

						# denominator
						log_Tij -= np.log10((pop_orig + sij_all)**alpha + 1)
						log_Tij -= np.log10((pop_orig + attractiveness + sij_all)**alpha +1)

						# scaling
						log_Tij += (np.log10(pop_oxfordshire) - np.log10(pop_traffic))

						# Linear OSM modification
						if len(OSM_values) != 0:
							# Linear modification
							#log_Tij += np.log10(sum( ci*vi/pop_dest for ci,vi in zip(c[1:],OSM_values) ))
							# Exponential OSM modification
							#log_Tij += np.log10(np.e)*sum( ci*vi/pop_dest for ci,vi in zip(c[1:],OSM_values) )
							# Syrian modification
							log_Tij += ( distance / sum( ci*vi/pop_dest for ci,vi in zip(c[1:],OSM_values) ))

						# Exponential OSM modification
						#log_Tij += np.log10(np.e)*sum( ci*vi/pop_dest for ci,vi in zip(c[1:],OSM_values) )

						# Apply cutoff
						m = min(log_y_data)
						#log_Tij = np.array([ i if i>m else m for i in log_Tij ])		

						global p1s, p2s
						p1s += [ list(params) + [0]*4 ]
						p2s += [ (log_Tij-log_y_data)**2 ]

						return log_Tij

				p0 = [0.0]*2 + [ 0.0 for _ in OSM_values ]
				p0 = np.random.rand(len(p0))
				func = rad_alpha2

				pops   = [ pop_oxfordshire, pop_traffic ]
				x_data = [alpha, pop_orig, pop_dest, distance, sij, Ti, pops ] + OSM_values + sij_OSM_values
				x_data = np.array([ np.array(x, dtype='float64') for x in x_data ])
				y_data = np.array(Tij.copy(), dtype='float64')
				log_y_data = np.log10(y_data)

				fitParams, fitCovariances = curve_fit(func, x_data, log_y_data, p0=p0, maxfev=50000)
				y_rad_alpha1 = func(x_data, *fitParams)

				#print('%d iterations for part 1' % len(p1s), end=', ')


				if with_sij_mod:
					def rad_alpha_mod2(inputs, *params):
						pop_orig, pop_dest, distance, sij, Ti, pops, y_rad_alpha = inputs[:7]
						pop_oxfordshire, pop_traffic = pops
						OSM_values = inputs[7:]

						A, B = params

						if (A<0) or (B<0):
							return -1e10

						else:
							# Non-modified model
							Tij = 10**y_rad_alpha

							# sij modification
							Tij *= ( A * np.exp(-sij/(B*100000)) + 1 )
							#Tij *= ( pop_orig ** C ) * ( pop_dest ** D )

							log_Tij = np.log10(Tij)
							global p1s, p2s
							p1s += [ p1s[-1][:-4] + list(params) ]
							p2s += [ ((log_Tij-log_y_data)**2)/sum(log_y_data) ]

							return Tij


					p0 = [0.0]*2
					p0 = np.random.rand(len(p0))
					func = rad_alpha_mod2

					pops   = [ pop_oxfordshire, pop_traffic ]
					x_data = [ pop_orig, pop_dest, distance, sij, Ti, pops, y_rad_alpha1 ] + OSM_values
					x_data = np.array([ np.array(x, dtype='float64') for x in x_data ])
					y_data = np.array(Tij.copy(), dtype='float64')
					log_y_data = np.log10(y_data)

					fitParams, fitCovariances = curve_fit(func, x_data, y_data, p0=p0, maxfev=30000)
					y_rad_alpha2 = func(x_data, *fitParams)

					y_rad_alpha2 = np.log10(y_rad_alpha2)

				else:
					y_rad_alpha2 = y_rad_alpha1

				#print('%d iterations for part 2' % len(p1s), end=', ')
                                
				NRMSE_final = sum(p2s[-1])
				#print('NRMSE = %g' % NRMSE_final)

				if NRMSE_final < min_NRMSE_final:
					y_final = y_rad_alpha2.copy()
					min_NRMSE_final = NRMSE_final
					print('iteration %d: NRMSE = %g' % (iterations, min_NRMSE_final) )

			except RuntimeError:
				print('Iteration %d, RuntimeError' % iterations)
				continue
					
		y_rad_alpha2 = y_final.copy()
		print('Final NRMSE = %g' % ( min_NRMSE_final) )		


		print("\nFitting Model 8: intervening opportunities model")
		# Model 8: intervening opportunities model

		min_NRMSE_final = np.infty
		y_final		 = np.ones_like(Tij)

		gc.collect()
		for iterations in range(N_FIT_ITERATIONS):
			try:
				p1s, p2s = [], []

				def inter_oppo(inputs,  *params):
					idx_OSM_values = 6
					idx_sij_OSM_values = idx_OSM_values + len(params[1:])

					pop_orig, pop_dest, distance, sij, Ti, pops = inputs[:idx_OSM_values]
					pop_oxfordshire, pop_traffic = pops   
					OSM_values	 = inputs[idx_OSM_values:idx_sij_OSM_values]
					sij_OSM_values = inputs[idx_sij_OSM_values:]

					c = params
					gamma = c[0]
					attractiveness = gamma
					sij_all		= sij

					if np.any(attractiveness<0) or np.any(sij_all<0):
						return -1e10
					else:
						# Linear OSM modification
						if len(OSM_values) != 0:
							#pop_dest_mod = pop_dest + sum( ci*vi/pop_dest for ci,vi in zip(c[1:],OSM_values) )
							# Linear modification
							#log_Tij += np.log10(sum( ci*vi/pop_dest for ci,vi in zip(c[1:],OSM_values) ))
							# Exponential OSM modification
							#log_Tij += np.log10(np.e)*sum( ci*vi/pop_dest for ci,vi in zip(c[1:],OSM_values) )
							# Syrian modification
							pop_dest_mod = pop_dest * np.exp( distance / sum( ci*vi/pop_dest for ci,vi in zip(c[1:],OSM_values) ))
						else:
							pop_dest_mod = pop_dest

						# Schneider's intervening opportunities
						Tij = Ti * ( np.exp(-gamma*sij_all) - np.exp(-gamma*(sij_all+pop_dest_mod)) )

						# scaling
						Tij *= (pop_oxfordshire/pop_traffic)

						# Exponential OSM modification
						#log_Tij += np.log10(np.e)*sum( ci*vi/pop_dest for ci,vi in zip(c[1:],OSM_values) )

						# Apply cutoff?
						#m = min(log_y_data)
						#log_Tij = np.array([ i if i>m else m for i in log_Tij ])		

						global p1s, p2s
						p1s += [ list(params) + [0]*4 ]
						p2s += [ (Tij-y_data)**2 ]

						return Tij

				p0 = [0.0] + [ 0.0 for _ in OSM_values ]
				p0 = np.random.rand(len(p0))
				func = inter_oppo

				pops   = [ pop_oxfordshire, pop_traffic ]
				x_data = [ pop_orig, pop_dest, distance, sij, Ti, pops ] + OSM_values + sij_OSM_values
				x_data = np.array([ np.array(x, dtype='float64') for x in x_data ])
				y_data = np.array(Tij.copy(), dtype='float64')
				log_y_data = np.log10(y_data)

				fitParams, fitCovariances = curve_fit(func, x_data, y_data, p0=p0, maxfev=50000)
				y_inter_oppo1 = func(x_data, *fitParams)

				#print('%d iterations for part 1' % len(p1s), end=', ')

				if with_sij_mod:
					def inter_oppo_mod(inputs, *params):
						pop_orig, pop_dest, distance, sij, Ti, pops, y_inter_oppo1 = inputs[:7]
						pop_oxfordshire, pop_traffic = pops
						OSM_values = inputs[7:]

						A, B = params

						if (A<0) or (B<0):
							return -1e10

						else:
							# Non-modified model
							Tij = y_inter_oppo1

							# sij modification
							Tij *= ( A * np.exp(-sij/(B*100000)) + 1 )
							#Tij *= ( pop_orig ** C ) * ( pop_dest ** D )

							#m = min(y_data)
							#Tij = np.array([ i if i>m else m for i in Tij ])	   

							log_Tij = np.log10(Tij)
							global p1s, p2s
							p1s += [ p1s[-1][:-4] + list(params) ]
							p2s += [ ((log_Tij-log_y_data)**2)/sum(log_y_data) ]

							return Tij


					p0 = [0.0]*2
					p0 = np.random.rand(len(p0))
					func = inter_oppo_mod

					pops   = [ pop_oxfordshire, pop_traffic ]
					x_data = [ pop_orig, pop_dest, distance, sij, Ti, pops, y_inter_oppo1 ] + OSM_values
					x_data = np.array([ np.array(x, dtype='float64') for x in x_data ])
					y_data = np.array(Tij.copy(), dtype='float64')

					fitParams, fitCovariances = curve_fit(func, x_data, y_data, p0=p0, maxfev=30000)
					y_inter_oppo = func(x_data, *fitParams)

					y_inter_oppo = np.log10(y_inter_oppo)

				else:
					y_inter_oppo = np.log10(y_inter_oppo1)

				#print('%d iterations for part 2' % len(p1s), end=', ')
				NRMSE_final = sum(p2s[-1])
				#print('NRMSE = %g' % NRMSE_final)

				if NRMSE_final < min_NRMSE_final:
					y_final = y_inter_oppo.copy()
					min_NRMSE_final = NRMSE_final
					print('iteration %d: NRMSE = %g' % (iterations, min_NRMSE_final) )

			except RuntimeError:
				print('Iteration %d, RuntimeError' % iterations)
				continue
					
		y_inter_oppo = y_final.copy()
		print('Final NRMSE = %g' % ( min_NRMSE_final) )

		#return y_data, log_y_data, y_rad_alpha3, y_inter_oppo

		# Save results

		from itertools import product

		def divide_by_rowsums(e):
			return e/e.sum(axis=1)[:,None]

		def scale_Tij(Tij_in, scale, population):
			tot_pop = sum(population)
			pop_scale = population/tot_pop
			
			if len(Tij_in.shape) == 1:
				# first make it a matrix
				L = int(0.5*(1+np.sqrt(1+4*len(Tij_in))))
				Tij = np.zeros((L,L))

				k = 0
				for i,j in product(range(L), range(L)):
					if i != j:
						Tij[i,j] = Tij_in[k]
						k += 1
			else:
				Tij = Tij_in.copy()
				
			Tij = divide_by_rowsums(Tij)
			#Tij *= pop_scale
			Tij = (pop_scale*Tij.T).T

			np.fill_diagonal(Tij, 0)

			Tij /= np.sum(Tij)
			Tij *= np.sum(scale)

			if len(Tij_in.shape) == 1:
				# remove diagonal entries again
				Tij_out = []
				for i,j in product(range(L), range(L)):
					if i != j:
						Tij_out.append(Tij[i,j])
			else:
				Tij_out = Tij
			return Tij_out		

		codes_in_order = []
		for origin in df_Tij_data.origin:
			if origin not in codes_in_order:
				codes_in_order += [ origin ]

		population = np.array([ df_population[ df_population.CODE==i ].population.values[0] for i in codes_in_order ])

		# Save everything

		df = df_Tij_data_filtered_2.copy()
		#df['meanvolume'] *= pop_oxfordshire/pop_traffic
		#df['Ti']		 *= pop_oxfordshire/pop_traffic

		# Make all trip interchange matrices

		all_df_Tijs = [ df.copy(), df.copy(), df.copy(), df.copy(), df.copy(), \
						df.copy(), df.copy(), df.copy(), df.copy()]

		cutoff_mask = df.meanvolume>0

		Tij_models = [ y_grav_pow, y_grav_exp, y_grav_pow2, y_grav_exp2, \
					  y_rad_ori, y_rad_alpha, y_rad_alpha2, y_inter_oppo ]

		for idx,Tijs in enumerate(Tij_models):
			trip_volumes = np.zeros_like(all_df_Tijs[0].meanvolume.values)
			trip_volumes[cutoff_mask] = 10**Tijs   
			all_df_Tijs[idx+1]['meanvolume'] = trip_volumes
		
				# Uncomment to do the scaling.
			# But remember it makes all fits bad.
			#all_df_Tijs[idx+1]['meanvolume'] = scale_Tij(trip_volumes, all_df_Tijs[0].meanvolume.values, population)
			

		all_df_names = [ 'data', 'grav_pow', 'grav_exp', 'grav_pow2', 'grav_exp2',
						'rad_ori', 'rad_alpha', 'rad_alpha_EPJ', 'inter_oppo' ]

		df_total = all_df_Tijs[0].copy()
		df_total = df_total[df_total['meanvolume'] > 0]
		del df_total['meanvolume']

		for i,df in enumerate(all_df_Tijs):
			df_total['meanvolume'+str(i)] = df[all_df_Tijs[0].meanvolume>0].meanvolume.values

		# Save predictions to file
		df = df_total.copy()

		df.columns = list(df_total.columns[:-9]) + [ 'Tij_data', \
													 'Tij_grav_pow', 'Tij_grav_exp',
													 'Tij_grav_pow2', 'Tij_grav_exp2',
													 'Tij_rad_ori', 'Tij_rad_alpha',
													 'Tij_rad_alpha_EPJ', 'Tij_inter_oppo' ]

		modification_tag = ''.join( str(i) for i in triple )
		if thresh2 == np.infty:
			print('data/traffic_predictions_with_'+modification_tag+'_Tij_above_%d.tsv' % thresh1)
			df.to_csv('data/traffic_predictions_with_'+modification_tag+'_Tij_above_%d.tsv' 
										  % (thresh1), sep='\t', index=False)
		else:
			df.to_csv('data/traffic_predictions_with_'+modification_tag+'_Tij_above_%d_below_%d.tsv' 
										  % (thresh1,thresh2), sep='\t', index=False)


# Common part of commuters:
def CPC(df_total, key0, key1):
	T0, T1 = df_total[key0].values, df_total[key1].values
	min_T1_T0 = np.minimum(T1,T0)  
	cpc = 2*sum(min_T1_T0)/( sum(T1) + sum(T0) )
	return cpc

# NRMSE
def NRMSE(df_total, key0, key1):
	#T0, T1 = df_total[key0].values, df_total[key1].values
	T0, T1 = np.log10(df_total[key0].values), np.log10(df_total[key1].values)
	return sum( (T1 - T0)**2 ) / sum(T0)

# Information gain statistic
def Igain(df_total, key0, key1, df_population):
	T0, T1 = df_total[key0].values, df_total[key1].values
	N = df_population.population.sum()
	logs = np.array([ np.log(i1/i0) if i1 != 0 else 0 for i1,i0 in zip(T1,T0) ])
	return sum( (T1/N) * logs )

# Common part of links:
def CPL(df_total, key0, key1):
	T0, T1 = df_total[key0].values, df_total[key1].values
	numerator = 2.0*sum((T1>0) & (T0>0))
	denominator   = sum(T1>0) + sum(T0>0)
	cpl = numerator/denominator
	return cpl

# Common part of commuters according to the distance
def CPCd(df_total, key0, key1):
	T0, T1 = df_total[key0].values, df_total[key1].values
	maxdist =  df_total.distance.max()
	bins = np.arange(0, maxdist+2, 2)
	
	# Nk is the number of individuals with a commuting distance between 2k-2 and 2k kms.
	Nk1, edges = np.histogram(df_total.distance, weights=df_total[key1], bins=bins)
	Nk0, edges = np.histogram(df_total.distance, weights=df_total[key0], bins=bins)

	cpcd = 2.0*sum(np.minimum(Nk1, Nk0))/( sum(Nk1) + sum(Nk0) )
	return cpcd

# Common Fraction of Commuters: like common part of commuters, but with every edge counting equally
def CFC(df_total, key0, key1):
	T0, T1 = df_total[key0].values, df_total[key1].values
	cfc = np.mean([ min(i/j, j/i) for i,j in zip(T0,T1) ])
	return cfc

# Common Fraction of Wards: like common part of commuters, but with every ward counting equally
def CFW(df_total, key0, key1, direction='origin'):
	T0, T1 = df_total[key0].values, df_total[key1].values
	allcodes = set(df_total.origin.values)
	
	commuters_per_ward_0 = []
	commuters_per_ward_1 = []
	for CODE in allcodes:
		df = df_total[df_total[direction]==CODE]
		commuters_per_ward_0 += [ df[key0].values.sum() ]
		commuters_per_ward_1 += [ df[key1].values.sum() ]
	commuters_per_ward_0 = np.array(commuters_per_ward_0)
	commuters_per_ward_1 = np.array(commuters_per_ward_1)
	
	min_cpw1_cpw0 = np.minimum(commuters_per_ward_0, commuters_per_ward_1)  
	cfw = 2*sum(min_cpw1_cpw0)/( sum(commuters_per_ward_0) + sum(commuters_per_ward_1) )
	return cfw

# Common part of commuters:
def CPC_with_bootstrap(df_total, key0, key1, seed=0, reps=10000):
	T0, T1 = df_total[key0].values, df_total[key1].values
	all_cpcs = []
	# do bootstrap
	np.random.seed(seed)
	r = np.arange(0,len(T0))
	rboots = np.random.choice(r, (len(r), reps), replace=True).T
	for k,rboot in enumerate(rboots):
		if k % 1000 == 0:
			print('Running bootstrap round %d for CPC' % (k))
		T0_new = T0[rboot]
		T1_new = T1[rboot]
		min_T1_T0 = np.minimum(T1_new,T0_new)  
		cpc = 2*sum(min_T1_T0)/( sum(T1_new) + sum(T0_new) )
		all_cpcs += [cpc]
	return all_cpcs

# Common Fraction of Commuters: like common part of commuters, but with every edge counting equally
def CFC_with_bootstrap(df_total, key0, key1, seed=0, reps=10000):
	T0, T1 = df_total[key0].values, df_total[key1].values
	all_cfcs = []
	# do bootstrap
	np.random.seed(seed)
	r = np.arange(0,len(T0))
	rboots = np.random.choice(r, (len(r), reps), replace=True).T
	print(rboots.shape)
	for k,rboot in enumerate(rboots):
		if k % 1000 == 0:
			print('Running bootstrap round %d for CFC' % (k))
		T0_new = T0[rboot]
		T1_new = T1[rboot]
		cfc = np.mean([ min(i/j, j/i) for i,j in zip(T0_new,T1_new) ])
		all_cfcs += [cfc]
	return all_cfcs


def calculate_gof(loaded_data):
	df_wards = loaded_data['df_wards']
	df_population = loaded_data['df_population']
	thresh1 = loaded_data['thresh1']
	thresh2 = loaded_data['thresh2']
	df_Tij_data = loaded_data['df_Tij_data'].copy()
	df = loaded_data['df_Tij_data_filtered'].copy()
	df_Tij_data_filtered_2 = loaded_data['df_Tij_data_filtered_2'].copy()

	thresh1 = 0
	from itertools import product
	l = range(2)
	for duple in product(l,l):
		triple = (duple[0],0,duple[1])

		modification_tag = ''.join(str(i) for i in triple)
		infile = 'data/traffic_predictions_with_'+modification_tag+'_Tij_above_%d.tsv' % (thresh1)
		df_total = pd.read_csv(infile, delimiter='\t')

		print('Read %s\n' % infile)

		model_keys = df_total.keys()[-9:]

		#df_T0 = all_df_Tijs[0]
		#colnames = [ 'modelname','CPC',r'$\log_{10}$(NRMSE)', 'I_gain','CPL','CPCd','CFC','CFWo', 'CFWd' ]
		colnames = [ 'modelname','CPC', 'NRMSE', 'I_gain','CPL','CPCd','CFC','CFWo', 'CFWd' ]

		key0 = model_keys[0]
		list_gof = []
		for idx, key1 in enumerate(model_keys[1:]):
			cpc   = CPC  (df_total, key0, key1)
			nrmse = NRMSE(df_total, key0, key1)
			igain = Igain(df_total, key0, key1, df_population)
			cpl   = CPL  (df_total, key0, key1)
			cpcd  = CPCd (df_total, key0, key1)
			cfc   = CFC(df_total, key0, key1)
			cfwo  = CFW(df_total, key0, key1, direction='origin')
			cfwd  = CFW(df_total, key0, key1, direction='destination')
			list_gof += [ (key1, cpc, nrmse, igain, cpl, cpcd, cfc, cfwo, cfwd ) ]

		df_gof = pd.DataFrame(list_gof, columns=colnames )
		df_gof.to_csv('data/models-with-'+modification_tag+'-gof-measures.csv')



# Bootstrap:
# https://stackoverflow.com/questions/15033511/compute-a-confidence-interval-from-sample-data
# or
# https://pypi.org/project/bootstrapped/
def calculate_gof_with_bootstrap(loaded_data, reps=10000):
	df_wards = loaded_data['df_wards']
	df_population = loaded_data['df_population']
	thresh1 = loaded_data['thresh1']
	thresh2 = loaded_data['thresh2']
	df_Tij_data = loaded_data['df_Tij_data'].copy()
	df = loaded_data['df_Tij_data_filtered'].copy()
	df_Tij_data_filtered_2 = loaded_data['df_Tij_data_filtered_2'].copy()

	thresh1 = 0
	from itertools import product
	l = range(2)
	for duple in product(l,l):
		triple = (duple[0],0,duple[1])
		modification_tag = ''.join(str(i) for i in triple)
		infile = 'data/traffic_predictions_with_'+modification_tag+'_Tij_above_%d.tsv' % (thresh1)
		df_total = pd.read_csv(infile, delimiter='\t')

		print('Read %s\n' % infile)

		model_keys = df_total.keys()[-9:]

		#df_T0 = all_df_Tijs[0]
		#colnames = [ 'modelname','CPC',r'$\log_{10}$(NRMSE)', 'I_gain','CPL','CPCd','CFC','CFWo', 'CFWd' ]
		colnames = [ 'modelname','NRMSE','I_gain','CPL','CPCd','CFWo', 'CFWd',
					 'CPC_mean','CPC_CI','CFC_mean','CFC_CI' ]

		key0 = model_keys[0]
		list_gof = []
		for idx, key1 in enumerate(model_keys[1:]):
			print('triple: '+modification_tag+', Model: '+key1)
			nrmse = 0 #NRMSE(df_total, key0, key1)
			igain = 0 #Igain(df_total, key0, key1, df_population)
			cpl   = 0 #CPL  (df_total, key0, key1)
			cpcd  = 0 #CPCd (df_total, key0, key1)
			cfwo  = 0 #CFW(df_total, key0, key1, direction='origin')
			cfwd  = 0 #CFW(df_total, key0, key1, direction='destination')

			cpcs  = CPC_with_bootstrap(df_total, key0, key1, seed=idx, reps=reps)
			cfcs  = CFC_with_bootstrap(df_total, key0, key1, seed=idx, reps=reps)

			a = cpcs
			cpc_mean  = np.mean(cpcs)
			cpc_CI    = st.t.interval(0.95, len(a)-1, loc=np.mean(a), scale=st.sem(a))
			outfile = 'data/CPC_with_{}_bootstraps_{}.csv'.format(reps,modification_tag)
			with open(outfile,'a') as f:
				f.write(key1+','+','.join(str(i) for i in a)+'\n')

			a = cfcs
			cfc_mean  = np.mean(cfcs)
			cfc_CI    = st.t.interval(0.95, len(a)-1, loc=np.mean(a), scale=st.sem(a))
			outfile = 'data/CFC_with_{}_bootstraps_{}.csv'.format(reps,modification_tag)
			with open(outfile,'a') as f:
				f.write(key1+','+','.join([str(i) for i in a])+'\n')

			list_gof += [ (key1, nrmse, igain, cpl, cpcd, cfwo, cfwd,
						   cpc_mean, cpc_CI, cfc_mean, cfc_CI ) ]

		df_gof = pd.DataFrame(list_gof, columns=colnames )
		df_gof.to_csv('data/models-with-'+modification_tag+'-gof-measures.csv')
		